/******************************************************************************
 * Copyright (c) 2010, Howard Butler
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following
 * conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in
 *       the documentation and/or other materials provided
 *       with the distribution.
 *     * Neither the name of the Martin Isenburg or Iowa Department
 *       of Natural Resources nor the names of its contributors may be
 *       used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 ****************************************************************************/

// This eliminates the need to include pdal base includes in code for the Util
// library.

#pragma once


#ifdef _WIN32
#   define PDAL_EXPORT   __declspec(dllexport)
#   define PDAL_DLL     PDAL_EXPORT
#   define PDAL_LOCAL
#   define PDAL_EXPORT_UNIX
#ifdef __GNUC__
#   define PDAL_EXPORT_DEPRECATED __attribute__((deprecated)) PDAL_EXPORT
#else
#   define PDAL_EXPORT_DEPRECATED   __declspec(deprecated, dllexport)
#endif
#else
#   define PDAL_EXPORT     __attribute__ ((visibility("default")))
// Keep the PDAL_DLL name around so any external plugins that still might have that defined
// can still compile and be used
#   define PDAL_DLL     PDAL_EXPORT
#   define PDAL_LOCAL   __attribute__((visibility("hidden")))
#   define PDAL_EXPORT_UNIX   __attribute__ ((visibility("default")))
#   define PDAL_EXPORT_DEPRECATED   __attribute__((deprecated))
#endif // _WIN32


