/******************************************************************************
* Copyright (c) 2022, Daniel Brookes (dbrookes@micromine.com)
*
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following
* conditions are met:
*
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in
*       the documentation and/or other materials provided
*       with the distribution.
*     * Neither the name of Hobu, Inc. nor the
*       names of its contributors may be used to endorse or promote
*       products derived from this software without specific prior
*       written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
* FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
* COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
* OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
* AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
* OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
* OF SUCH DAMAGE.
****************************************************************************/

#pragma once

#include "pdal/Reader.hpp"

namespace pdal
{

class PDAL_EXPORT PtxReader : public Reader
{
public:
    PtxReader() = default;
    ~PtxReader();
    virtual std::string getName() const override;

protected:
    struct PtxHeader
    {
        size_t m_columns;
        size_t m_rows;
        std::array<double, 16> m_transform;

        void applyTransform(double &x, double &y, double &z) const;
        PtxHeader() : m_columns(0), m_rows(0) {};
    };

    PtxHeader readHeader();

private:
    virtual void initialize(PointTableRef table) override;
    virtual void addArgs(ProgramArgs& args) override;
    virtual void addDimensions(PointLayoutPtr layout) override;
    virtual void ready(PointTableRef table) override;
    virtual point_count_t read(PointViewPtr view,
                               point_count_t numPts) override;
    virtual void done(PointTableRef table) override;

private:
    bool m_discardMissingPoints{ true };
    std::istream* m_istream{ nullptr };
    Dimension::IdList m_dimensions;
};

} // namespace pdal
